@@ -35,12 +35,14 @@
 
 (require 'tramp)
 
+;;;###tramp-autoload
 (defcustom tramp-adb-program "adb"
   "Name of the Android Debug Bridge program."
   :group 'tramp
   :version "24.4"
   :type 'string)
 
+;;;###tramp-autoload
 (defcustom tramp-adb-connect-if-not-connected nil
   "Try to run `adb connect' if provided device is not connected currently.
 It is used for TCP/IP devices."
@@ -52,10 +54,11 @@ It is used for TCP/IP devices."
 (defconst tramp-adb-method "adb"
   "When this method name is used, forward all calls to Android Debug Bridge.")
 
+;;;###tramp-autoload
 (defcustom tramp-adb-prompt
   "^[[:digit:]]*|?[[:alnum:]\e;[]*@?[[:alnum:]]*[^#\\$]*[#\\$][[:space:]]"
   "Regexp used as prompt in almquist shell."
-  :type 'string
+  :type 'regexp
   :version "24.4"
   :group 'tramp)
 
@@ -91,7 +94,7 @@ It is used for TCP/IP devices."
   '((access-file . tramp-handle-access-file)
     (add-name-to-file . tramp-handle-add-name-to-file)
     ;; `byte-compiler-base-file-name' performed by default handler.
-    ;; `copy-directory' performed by default handler.
+    (copy-directory . tramp-handle-copy-directory)
     (copy-file . tramp-adb-handle-copy-file)
     (delete-directory . tramp-adb-handle-delete-directory)
     (delete-file . tramp-adb-handle-delete-file)
@@ -168,20 +171,20 @@ It is used for TCP/IP devices."
 ;; tramp-loaddefs.el.  Otherwise, there would be recursive autoloading.
 ;;;###tramp-autoload
 (defsubst tramp-adb-file-name-p (filename)
-  "Check if it's a filename for ADB."
+  "Check if it's a FILENAME for ADB."
   (and (tramp-tramp-file-p filename)
        (string= (tramp-file-name-method (tramp-dissect-file-name filename))
 		tramp-adb-method)))
 
 ;;;###tramp-autoload
-(defun tramp-adb-file-name-handler (operation &rest args)
+(defun tramp-adb-file-name-handler (operation &rest arguments)
   "Invoke the ADB handler for OPERATION.
-First arg specifies the OPERATION, second arg is a list of arguments to
-pass to the OPERATION."
+First arg specifies the OPERATION, second arg is a list of
+ARGUMENTS to pass to the OPERATION."
   (let ((fn (assoc operation tramp-adb-file-name-handler-alist)))
     (if fn
-	(save-match-data (apply (cdr fn) args))
-      (tramp-run-real-handler operation args))))
+	(save-match-data (apply (cdr fn) arguments))
+      (tramp-run-real-handler operation arguments))))
 
 ;;;###tramp-autoload
 (tramp--with-startup
@@ -229,96 +232,100 @@ pass to the OPERATION."
 ;; code could be shared?
 (defun tramp-adb-handle-file-truename (filename)
   "Like `file-truename' for Tramp files."
-   ;; Preserve trailing "/".
+  ;; Preserve trailing "/".
   (funcall
-   (if (string-equal (file-name-nondirectory filename) "")
+   (if (tramp-compat-directory-name-p filename)
        #'file-name-as-directory #'identity)
-   (with-parsed-tramp-file-name (expand-file-name filename) nil
-     (tramp-make-tramp-file-name
-      v
-      (with-tramp-file-property v localname "file-truename"
-	(let ((result nil)			; result steps in reverse order
-	      (quoted (tramp-compat-file-name-quoted-p localname)))
-	  (tramp-message v 4 "Finding true name for `%s'" filename)
-	  (let* ((steps (split-string localname "/" 'omit))
-		 (localnamedir (tramp-run-real-handler
-				'file-name-as-directory (list localname)))
-		 (is-dir (string= localname localnamedir))
-		 (thisstep nil)
-		 (numchase 0)
-		 ;; Don't make the following value larger than
-		 ;; necessary.  People expect an error message in a
-		 ;; timely fashion when something is wrong; otherwise
-		 ;; they might think that Emacs is hung.  Of course,
-		 ;; correctness has to come first.
-		 (numchase-limit 20)
-		 symlink-target)
-	    (while (and steps (< numchase numchase-limit))
-	      (setq thisstep (pop steps))
-	      (tramp-message
-	       v 5 "Check %s"
-	       (string-join
-		(append '("") (reverse result) (list thisstep)) "/"))
-	      (setq symlink-target
-		    (tramp-compat-file-attribute-type
-		     (file-attributes
-		      (tramp-make-tramp-file-name
-		       v
-		       (string-join
-			(append '("") (reverse result) (list thisstep)) "/")))))
-	      (cond ((string= "." thisstep)
-		     (tramp-message v 5 "Ignoring step `.'"))
-		    ((string= ".." thisstep)
-		     (tramp-message v 5 "Processing step `..'")
-		     (pop result))
-		    ((stringp symlink-target)
-		     ;; It's a symlink, follow it.
-		     (tramp-message v 5 "Follow symlink to %s" symlink-target)
-		     (setq numchase (1+ numchase))
-		     (when (file-name-absolute-p symlink-target)
-		       (setq result nil))
-		     ;; If the symlink was absolute, we'll get a string
-		     ;; like "/user@host:/some/target"; extract the
-		     ;; "/some/target" part from it.
-		     (when (tramp-tramp-file-p symlink-target)
-		       (unless (tramp-equal-remote filename symlink-target)
-			 (tramp-error
-			  v 'file-error
-			  "Symlink target `%s' on wrong host" symlink-target))
-		       (setq symlink-target localname))
-		     (setq steps
-			   (append (split-string symlink-target "/" 'omit)
-				   steps)))
-		    (t
-		     ;; It's a file.
-		     (setq result (cons thisstep result)))))
-	    (when (>= numchase numchase-limit)
-	      (tramp-error
-	       v 'file-error
-	       "Maximum number (%d) of symlinks exceeded" numchase-limit))
-	    (setq result (reverse result))
-	    ;; Combine list to form string.
-	    (setq result
-		  (if result
-		      (string-join (cons "" result) "/")
-		    "/"))
-	    (when (and is-dir (or (string-empty-p result)
-				  (not (string= (substring result -1) "/"))))
-	      (setq result (concat result "/"))))
+   ;; Quote properly.
+   (funcall
+    (if (tramp-compat-file-name-quoted-p filename)
+	#'tramp-compat-file-name-quote #'identity)
+    (with-parsed-tramp-file-name
+	(tramp-compat-file-name-unquote (expand-file-name filename)) nil
+      (tramp-make-tramp-file-name
+       v
+       (with-tramp-file-property v localname "file-truename"
+	 (let (result)			; result steps in reverse order
+	   (tramp-message v 4 "Finding true name for `%s'" filename)
+	   (let* ((steps (split-string localname "/" 'omit))
+		  (localnamedir (tramp-run-real-handler
+				 'file-name-as-directory (list localname)))
+		  (is-dir (string= localname localnamedir))
+		  (thisstep nil)
+		  (numchase 0)
+		  ;; Don't make the following value larger than
+		  ;; necessary.  People expect an error message in a
+		  ;; timely fashion when something is wrong; otherwise
+		  ;; they might think that Emacs is hung.  Of course,
+		  ;; correctness has to come first.
+		  (numchase-limit 20)
+		  symlink-target)
+	     (while (and steps (< numchase numchase-limit))
+	       (setq thisstep (pop steps))
+	       (tramp-message
+		v 5 "Check %s"
+		(string-join
+		 (append '("") (reverse result) (list thisstep)) "/"))
+	       (setq symlink-target
+		     (tramp-compat-file-attribute-type
+		      (file-attributes
+		       (tramp-make-tramp-file-name
+			v
+			(string-join
+			 (append
+			  '("") (reverse result) (list thisstep)) "/")))))
+	       (cond ((string= "." thisstep)
+		      (tramp-message v 5 "Ignoring step `.'"))
+		     ((string= ".." thisstep)
+		      (tramp-message v 5 "Processing step `..'")
+		      (pop result))
+		     ((stringp symlink-target)
+		      ;; It's a symlink, follow it.
+		      (tramp-message v 5 "Follow symlink to %s" symlink-target)
+		      (setq numchase (1+ numchase))
+		      (when (file-name-absolute-p symlink-target)
+			(setq result nil))
+		      ;; If the symlink was absolute, we'll get a string
+		      ;; like "/user@host:/some/target"; extract the
+		      ;; "/some/target" part from it.
+		      (when (tramp-tramp-file-p symlink-target)
+			(unless (tramp-equal-remote filename symlink-target)
+			  (tramp-error
+			   v 'file-error
+			   "Symlink target `%s' on wrong host" symlink-target))
+			(setq symlink-target localname))
+		      (setq steps
+			    (append (split-string symlink-target "/" 'omit)
+				    steps)))
+		     (t
+		      ;; It's a file.
+		      (setq result (cons thisstep result)))))
+	     (when (>= numchase numchase-limit)
+	       (tramp-error
+		v 'file-error
+		"Maximum number (%d) of symlinks exceeded" numchase-limit))
+	     (setq result (reverse result))
+	     ;; Combine list to form string.
+	     (setq result
+		   (if result
+		       (string-join (cons "" result) "/")
+		     "/"))
+	     (when (and is-dir (or (string-empty-p result)
+				   (not (string= (substring result -1) "/"))))
+	       (setq result (concat result "/"))))
 
-	  ;; Detect cycle.
-	  (when (and (file-symlink-p filename)
-		     (string-equal result localname))
-	    (tramp-error
-	     v 'file-error
-	     "Apparent cycle of symbolic links for %s" filename))
-	  ;; If the resulting localname looks remote, we must quote it
-	  ;; for security reasons.
-	  (when (or quoted (file-remote-p result))
-	    (let (file-name-handler-alist)
-	      (setq result (tramp-compat-file-name-quote result))))
-	  (tramp-message v 4 "True name of `%s' is `%s'" localname result)
-	  result))))))
+	   ;; Detect cycle.
+	   (when (and (file-symlink-p filename)
+		      (string-equal result localname))
+	     (tramp-error
+	      v 'file-error
+	      "Apparent cycle of symbolic links for %s" filename))
+	   ;; If the resulting localname looks remote, we must quote it
+	   ;; for security reasons.
+	   (when (file-remote-p result)
+	     (setq result (tramp-compat-file-name-quote result 'top)))
+	   (tramp-message v 4 "True name of `%s' is `%s'" localname result)
+	   result)))))))
 
 (defun tramp-adb-handle-file-attributes (filename &optional id-format)
   "Like `file-attributes' for Tramp files."
@@ -376,6 +383,10 @@ pass to the OPERATION."
 (defun tramp-adb-handle-directory-files-and-attributes
   (directory &optional full match nosort id-format)
   "Like `directory-files-and-attributes' for Tramp files."
+  (unless (file-exists-p directory)
+    (tramp-error
+     (tramp-dissect-file-name directory) tramp-file-missing
+     "No such file or directory" directory))
   (when (file-directory-p directory)
     (with-parsed-tramp-file-name (expand-file-name directory) nil
       (copy-tree
@@ -455,7 +466,7 @@ Convert (\"-al\") to (\"-a\" \"-l\").  Remove arguments like \"--dired\"."
 
 (defun tramp-adb-sh-fix-ls-output (&optional sort-by-time)
   "Insert dummy 0 in empty size columns.
-Androids \"ls\" command doesn't insert size column for directories:
+Android's \"ls\" command doesn't insert size column for directories:
 Emacs dired can't find files."
   (save-excursion
     ;; Insert missing size.
@@ -503,11 +514,12 @@ Emacs dired can't find files."
   "Like `make-directory' for Tramp files."
   (setq dir (expand-file-name dir))
   (with-parsed-tramp-file-name dir nil
+    (when (and (null parents) (file-exists-p dir))
+      (tramp-error v 'file-already-exists "Directory already exists %s" dir))
     (when parents
       (let ((par (expand-file-name ".." dir)))
 	(unless (file-directory-p par)
 	  (make-directory par parents))))
-    (tramp-flush-file-properties v (file-name-directory localname))
     (tramp-flush-directory-properties v localname)
     (unless (or (tramp-adb-send-command-and-check
 		 v (format "mkdir %s" (tramp-shell-quote-argument localname)))
@@ -518,10 +530,8 @@ Emacs dired can't find files."
   "Like `delete-directory' for Tramp files."
   (setq directory (expand-file-name directory))
   (with-parsed-tramp-file-name (file-truename directory) nil
-    (tramp-flush-file-properties v (file-name-directory localname))
     (tramp-flush-directory-properties v localname))
   (with-parsed-tramp-file-name directory nil
-    (tramp-flush-file-properties v (file-name-directory localname))
     (tramp-flush-directory-properties v localname)
     (tramp-adb-barf-unless-okay
      v (format "%s %s"
@@ -533,7 +543,6 @@ Emacs dired can't find files."
   "Like `delete-file' for Tramp files."
   (setq filename (expand-file-name filename))
   (with-parsed-tramp-file-name filename nil
-    (tramp-flush-file-properties v (file-name-directory localname))
     (tramp-flush-file-properties v localname)
     (tramp-adb-barf-unless-okay
      v (format "rm %s" (tramp-shell-quote-argument localname))
@@ -624,7 +633,6 @@ But handle the case, if the \"test\" command is not available."
 
     ;; We must also flush the cache of the directory, because
     ;; `file-attributes' reads the values from there.
-    (tramp-flush-file-properties v (file-name-directory localname))
     (tramp-flush-file-properties v localname)
     (let* ((curbuf (current-buffer))
 	   (tmpfile (tramp-compat-make-temp-file filename)))
@@ -662,14 +670,12 @@ But handle the case, if the \"test\" command is not available."
 (defun tramp-adb-handle-set-file-modes (filename mode)
   "Like `set-file-modes' for Tramp files."
   (with-parsed-tramp-file-name filename nil
-    (tramp-flush-file-properties v (file-name-directory localname))
     (tramp-flush-file-properties v localname)
     (tramp-adb-send-command-and-check v (format "chmod %o %s" mode localname))))
 
 (defun tramp-adb-handle-set-file-times (filename &optional time)
   "Like `set-file-times' for Tramp files."
   (with-parsed-tramp-file-name filename nil
-    (tramp-flush-file-properties v (file-name-directory localname))
     (tramp-flush-file-properties v localname)
     (let ((time (if (or (null time)
 			(tramp-compat-time-equal-p time tramp-time-doesnt-exist)
@@ -706,19 +712,24 @@ PRESERVE-UID-GID and PRESERVE-EXTENDED-ATTRIBUTES are completely ignored."
     (let ((t1 (tramp-tramp-file-p filename))
 	  (t2 (tramp-tramp-file-p newname)))
       (with-parsed-tramp-file-name (if t1 filename newname) nil
+	(unless (file-exists-p filename)
+	  (tramp-error
+	   v tramp-file-missing
+	   "Copying file" "No such file or directory" filename))
+	(when (and (not ok-if-already-exists) (file-exists-p newname))
+	  (tramp-error v 'file-already-exists newname))
+	(when (and (file-directory-p newname)
+		   (not (tramp-compat-directory-name-p newname)))
+	  (tramp-error v 'file-error "File is a directory %s" newname))
+
 	(with-tramp-progress-reporter
 	    v 0 (format "Copying %s to %s" filename newname)
-
 	  (if (and t1 t2 (tramp-equal-remote filename newname))
 	      (let ((l1 (tramp-compat-file-local-name filename))
 		    (l2 (tramp-compat-file-local-name newname)))
-		(when (and (not ok-if-already-exists)
-			   (file-exists-p newname))
-		  (tramp-error v 'file-already-exists newname))
 		;; We must also flush the cache of the directory,
 		;; because `file-attributes' reads the values from
 		;; there.
-		(tramp-flush-file-properties v (file-name-directory l2))
 		(tramp-flush-file-properties v l2)
 		;; Short track.
 		(tramp-adb-barf-unless-okay
@@ -753,8 +764,6 @@ PRESERVE-UID-GID and PRESERVE-EXTENDED-ATTRIBUTES are completely ignored."
 		  ;; We must also flush the cache of the directory,
 		  ;; because `file-attributes' reads the values from
 		  ;; there.
-		  (tramp-flush-file-properties
-		   v (file-name-directory localname))
 		  (tramp-flush-file-properties v localname)
 		  (when (tramp-adb-execute-adb-command
 			 v "push"
@@ -785,22 +794,26 @@ PRESERVE-UID-GID and PRESERVE-EXTENDED-ATTRIBUTES are completely ignored."
     (let ((t1 (tramp-tramp-file-p filename))
 	  (t2 (tramp-tramp-file-p newname)))
       (with-parsed-tramp-file-name (if t1 filename newname) nil
+	(unless (file-exists-p filename)
+	  (tramp-error
+	   v tramp-file-missing
+	   "Renaming file" "No such file or directory" filename))
+	(when (and (not ok-if-already-exists) (file-exists-p newname))
+	  (tramp-error v 'file-already-exists newname))
+	(when (and (file-directory-p newname)
+		   (not (tramp-compat-directory-name-p newname)))
+	  (tramp-error v 'file-error "File is a directory %s" newname))
+
 	(with-tramp-progress-reporter
 	    v 0 (format "Renaming %s to %s" filename newname)
-
 	  (if (and t1 t2
 		   (tramp-equal-remote filename newname)
 		   (not (file-directory-p filename)))
 	      (let ((l1 (tramp-compat-file-local-name filename))
 		    (l2 (tramp-compat-file-local-name newname)))
-		(when (and (not ok-if-already-exists)
-			   (file-exists-p newname))
-		  (tramp-error v 'file-already-exists newname))
 		;; We must also flush the cache of the directory, because
 		;; `file-attributes' reads the values from there.
-		(tramp-flush-file-properties v (file-name-directory l1))
 		(tramp-flush-file-properties v l1)
-		(tramp-flush-file-properties v (file-name-directory l2))
 		(tramp-flush-file-properties v l2)
 		;; Short track.
 		(tramp-adb-barf-unless-okay
@@ -1090,7 +1103,7 @@ E.g. a host name \"192.168.1.1#5555\" returns \"192.168.1.1:5555\"
 		 vec 'file-error "Could not find device %s" host)))))))
 
 (defun tramp-adb-execute-adb-command (vec &rest args)
-  "Returns nil on success error-output on failure."
+  "Return nil on success error-output on failure."
   (when (and (> (length (tramp-file-name-host vec)) 0)
 	     ;; The -s switch is only available for ADB device commands.
 	     (not (member (car args) '("connect" "disconnect"))))
@@ -1104,7 +1117,7 @@ E.g. a host name \"192.168.1.1#5555\" returns \"192.168.1.1:5555\"
       (tramp-message vec 6 "%s" (buffer-string)))))
 
 (defun tramp-adb-find-test-command (vec)
-  "Checks, whether the ash has a builtin \"test\" command.
+  "Check whether the ash has a builtin \"test\" command.
 This happens for Android >= 4.0."
   (with-tramp-connection-property vec "test"
     (tramp-adb-send-command-and-check vec "type test")))
@@ -1143,8 +1156,7 @@ the exit status is not equal 0, and t otherwise."
 	   (format "%s; echo tramp_exit_status $?" command)
 	 "echo tramp_exit_status $?"))
   (with-current-buffer (tramp-get-connection-buffer vec)
-    (goto-char (point-max))
-    (unless (re-search-backward "tramp_exit_status [0-9]+" nil t)
+    (unless (tramp-search-regexp "tramp_exit_status [0-9]+")
       (tramp-error
        vec 'file-error "Couldn't find exit status of `%s'" command))
     (skip-chars-forward "^ ")
@@ -1178,9 +1190,8 @@ FMT and ARGS are passed to `error'."
             (when (re-search-forward prompt (point-at-eol) t)
               (forward-line 1)
               (delete-region (point-min) (point)))
-	    (goto-char (point-max))
-	    (re-search-backward prompt nil t)
-	    (delete-region (point) (point-max)))
+	    (when (tramp-search-regexp prompt)
+	      (delete-region (point) (point-max))))
 	(if timeout
 	    (tramp-error
 	     proc 'file-error
@@ -1192,6 +1203,10 @@ FMT and ARGS are passed to `error'."
   "Maybe open a connection VEC.
 Does not do anything if a connection is already open, but re-opens the
 connection if a previous connection has died for some reason."
+  ;; During completion, don't reopen a new connection.
+  (unless (tramp-connectable-p vec)
+    (throw 'non-essential 'non-essential))
+
   (let* ((buf (tramp-get-connection-buffer vec))
 	 (p (get-buffer-process buf))
 	 (host (tramp-file-name-host vec))
@@ -1205,14 +1220,6 @@ connection if a previous connection has died for some reason."
       (tramp-error vec 'file-error "Cannot switch to user `%s'" user))
 
     (unless (process-live-p p)
-      ;; During completion, don't reopen a new connection.  We check
-      ;; this for the process related to `tramp-buffer-name';
-      ;; otherwise `start-file-process' wouldn't run ever when
-      ;; `non-essential' is non-nil.
-      (when (and (tramp-completion-mode-p)
-		 (null (get-process (tramp-buffer-name vec))))
-	(throw 'non-essential 'non-essential))
-
       (save-match-data
 	(when (and p (processp p)) (delete-process p))
 	(if (zerop (length device))
