@@ -31,11 +31,21 @@
 
 ;; Pacify byte-compiler.
 (require 'cl-lib)
+(declare-function info-lookup->cache "info-look")
+(declare-function info-lookup->mode-cache "info-look")
+(declare-function info-lookup->mode-value "info-look")
+(declare-function info-lookup->other-modes "info-look")
+(declare-function info-lookup->topic-cache "info-look")
+(declare-function info-lookup->topic-value "info-look")
+(declare-function info-lookup-maybe-add-help "info-look")
 (declare-function recentf-cleanup "recentf")
 (declare-function tramp-dissect-file-name "tramp")
 (declare-function tramp-file-name-equal-p "tramp")
 (declare-function tramp-tramp-file-p "tramp")
 (defvar eshell-path-env)
+(defvar ido-read-file-name-non-ido)
+(defvar info-lookup-alist)
+(defvar ivy-completing-read-handlers-alist)
 (defvar recentf-exclude)
 (defvar tramp-current-connection)
 (defvar tramp-postfix-host-format)
@@ -69,6 +79,7 @@ special handling of `substitute-in-file-name'."
 			 #'tramp-rfn-eshadow-setup-minibuffer)))
 
 (defun tramp-rfn-eshadow-update-overlay-regexp ()
+  "An overlay covering the shadowed part of the filename."
   (format "[^%s/~]*\\(/\\|~\\)" tramp-postfix-host-format))
 
 ;; Package rfn-eshadow is preloaded in Emacs, but for some reason,
@@ -169,6 +180,87 @@ NAME must be equal to `tramp-current-connection'."
 	      (remove-hook 'tramp-cleanup-all-connections-hook
 			   #'tramp-recentf-cleanup-all))))
 
+;;; Integration of ido.el:
+
+(with-eval-after-load 'ido
+  (add-to-list 'ido-read-file-name-non-ido 'tramp-rename-files)
+  (add-to-list 'ido-read-file-name-non-ido 'tramp-these-rename-files)
+  (add-hook 'tramp-integration-unload-hook
+	    (lambda ()
+	      (setq ido-read-file-name-non-ido
+		    (delq 'tramp-these-rename-files ido-read-file-name-non-ido)
+		    ido-read-file-name-non-ido
+		    (delq 'tramp-rename-files ido-read-file-name-non-ido)))))
+
+;;; Integration of ivy.el:
+
+(with-eval-after-load 'ivy
+  (add-to-list 'ivy-completing-read-handlers-alist
+	       '(tramp-rename-files . completing-read-default))
+  (add-to-list 'ivy-completing-read-handlers-alist
+	       '(tramp-these-rename-files . completing-read-default))
+  (add-hook
+   'tramp-integration-unload-hook
+   (lambda ()
+     (setq ivy-completing-read-handlers-alist
+	   (delete
+	    (assq 'tramp-these-rename-files ivy-completing-read-handlers-alist)
+	    ivy-completing-read-handlers-alist)
+	   ivy-completing-read-handlers-alist
+	   (delete
+	    (assq 'tramp-rename-files ivy-completing-read-handlers-alist)
+	    ivy-completing-read-handlers-alist)))))
+
+;;; Integration of info-look.el:
+
+(with-eval-after-load 'info-look
+  ;; Create a pseudo mode `tramp-info-lookup-mode' for Tramp symbol lookup.
+  (info-lookup-maybe-add-help
+   :mode 'tramp-info-lookup-mode :topic 'symbol
+   :regexp "[^][()`'‘’,\" \t\n]+"
+   :doc-spec '(("(tramp)Function Index" nil "^ -+ .*: " "\\( \\|$\\)")
+	       ("(tramp)Variable Index" nil "^ -+ .*: " "\\( \\|$\\)")))
+
+  (add-hook
+   'tramp-integration-unload-hook
+   (lambda ()
+     (setcdr (assq 'symbol info-lookup-alist)
+	     (delete (info-lookup->mode-value 'symbol 'tramp-info-lookup-mode)
+		     (info-lookup->topic-value 'symbol)))
+     (setcdr (info-lookup->cache 'symbol)
+	     (delete (info-lookup->mode-cache 'symbol 'tramp-info-lookup-mode)
+		     (info-lookup->topic-cache 'symbol)))))
+
+  (dolist (mode (mapcar 'car (info-lookup->topic-value 'symbol)))
+    ;; Add `tramp-info-lookup-mode' to `other-modes' for either
+    ;; `emacs-lisp-mode' itself, or to modes which use
+    ;; `emacs-lisp-mode' as `other-modes'.  Reset `info-lookup-cache'.
+    (when (and (or (equal mode 'emacs-lisp-mode)
+		   (memq
+		    'emacs-lisp-mode (info-lookup->other-modes 'symbol mode)))
+	       (not (memq 'tramp-info-lookup-mode
+			  (info-lookup->other-modes 'symbol mode))))
+      (setcdr (info-lookup->mode-value 'symbol mode)
+	      (append (butlast (cdr (info-lookup->mode-value 'symbol mode)))
+		      `((tramp-info-lookup-mode
+			 . ,(info-lookup->other-modes 'symbol mode)))))
+      (setcdr (info-lookup->cache 'symbol)
+	      (delete (info-lookup->mode-cache 'symbol mode)
+		      (info-lookup->topic-cache 'symbol)))
+
+      (add-hook
+       'tramp-integration-unload-hook
+       `(lambda ()
+	  (setcdr (info-lookup->mode-value 'symbol ',mode)
+		  (append (butlast
+			   (cdr (info-lookup->mode-value 'symbol ',mode)))
+			  (list
+			   (delq 'tramp-info-lookup-mode
+				 (info-lookup->other-modes 'symbol ',mode)))))
+	  (setcdr (info-lookup->cache 'symbol)
+		  (delete (info-lookup->mode-cache 'symbol ',mode)
+			  (info-lookup->topic-cache 'symbol))))))))
+
 ;;; Default connection-local variables for Tramp:
 
 (defconst tramp-connection-local-default-profile
