@@ -17,6 +17,10 @@
 ;; You should have received a copy of the GNU General Public License
 ;; along with this program.  If not, see `https://www.gnu.org/licenses/'.
 
+;;; Commentary:
+
+;; A testsuite for testing file archives.
+
 ;;; Code:
 
 ;; The `tramp-archive-testnn-*' tests correspond to the respective
@@ -54,11 +58,11 @@
   "A directory file name, which looks like an archive.")
 
 (setq password-cache-expiry nil
-      tramp-verbose 0
       tramp-cache-read-persistent-data t ;; For auth-sources.
       tramp-copy-size-limit nil
       tramp-message-show-message nil
-      tramp-persistency-file-name nil)
+      tramp-persistency-file-name nil
+      tramp-verbose 0)
 
 (defun tramp-archive--test-make-temp-name ()
   "Return a temporary file name for test.
@@ -569,7 +573,7 @@ This checks also `file-name-as-directory', `file-name-directory',
 	     (looking-at-p
 	      (concat
 	       ;; There might be a summary line.
-	       "\\(total.+[[:digit:]]+\n\\)?"
+	       "\\(total.+[[:digit:]]+ ?[kKMGTPEZY]?i?B?\n\\)?"
 	       ;; We don't know in which order the files appear.
 	       (format
 		"\\(.+ %s\\( ->.+\\)?\n\\)\\{%d\\}"
@@ -959,7 +963,8 @@ This tests also `file-executable-p', `file-writable-p' and `set-file-modes'."
     (tramp-archive-cleanup-hash)))
 
 (defun tramp-archive-test-all (&optional interactive)
-  "Run all tests for \\[tramp-archive]."
+  "Run all tests for \\[tramp-archive].
+If INTERACTIVE is non-nil, the tests are run interactively."
   (interactive "p")
   (funcall
    (if interactive #'ert-run-tests-interactively #'ert-run-tests-batch)
